@@ -1,10 +1,16 @@
-import shelve, pickle, cx_Oracle, datetime, sys, itertools
+import shelve, pickle, datetime, sys, itertools
+from sqlpython import cx_Oracle, psycopg2
 shelvename = 'plot.shelve'
 
 try:
     import pylab
     class Plot(object):
-        plottable_types = (cx_Oracle.NUMBER, datetime.datetime)    
+        plottable_types = [cx_Oracle and cx_Oracle.NUMBER, 
+                           datetime.datetime]
+        if psycopg2:
+            for val in psycopg2.NUMBER.values:
+                plottable_types.append(val)
+        #TODO: add non-Oracle types
         def __init__(self):
             self.legends = []
             self.yserieslists = []
@@ -76,7 +82,7 @@ try:
                 self.bar()
             pylab.xlabel(self.xlabel)
             pylab.title(self.title)
-            pylab.legend([p[0] for p in self.yplots], self.legends, shadow=True)
+            pylab.legend([p[0] for p in self.yplots], self.legends, shadow=True) #dies for psql on pies
             pylab.show()
             print 'You can edit this plot from the command prompt (outside sqlpython) by running'
             print "ipython -pylab -c 'import sqlpython.plothandler; sqlpython.plothandler.Plot().unshelve()'"
