@@ -27,6 +27,7 @@ import sys, os, re, sqlpython, pyparsing, re, completion
 import datetime, pickle, binascii, subprocess, time, itertools, hashlib
 import traceback, operator
 from cmd2 import Cmd, make_option, options, Statekeeper, Cmd2TestCase, options_defined
+import operator
 from output_templates import output_templates
 from metadata import metaqueries
 from plothandler import Plot
@@ -162,6 +163,10 @@ class SoftwareSearcher(object):
 Looked for these programs:
 %s""" % (self.purpose, __file__, "\n".join([s[0] for s in self.softwareList])))
 
+class Dummy_Options(object):
+    all = True
+dummy_options = Dummy_Options()
+
 softwareLists = {
     'diff/merge': [  
         ('/usr/bin/meld',"%s %s %s"),
@@ -587,17 +592,12 @@ class sqlpyPlus(sqlpython.sqlpython):
         r'select\s+(.*)from',
         re.IGNORECASE | re.DOTALL | re.MULTILINE)        
     def completedefault(self, text, line, begidx, endidx):
-        (username, gerald_schema) = self.metadata()
-        segment = completion.whichSegment(line)        
-        text = text.upper()
+        segment = completion.whichSegment(line)       
         if segment in ('select', 'where', 'having', 'set', 'order by', 'group by'):
-            completions = [c for c in schemas[username].column_names if c.startswith(text)] \
-                          or [c for c in schemas.qual_column_names if c.startswith(text)]
-                          # TODO: the latter not working
+            completions = [c[-1] for c in self.current_instance.columns(text + '%', '%', dummy_options)]
         elif segment in ('from', 'update', 'insert into'):
-            # print schemas[username].table_names
-            # TODO: from postgres, these table names are jrrt.fishies, etc.
-            completions = [t for t in schemas[username].table_names if t.startswith(text)]
+            #print self.current_instance.tables_and_views(text)
+            completions = [t[0] for t in self.current_instance.tables_and_views(text)]
         elif segment == 'beginning':
             completions = [n for n in self.get_names() if n.startswith('do_')] + [
                            'insert', 'update', 'delete', 'drop', 'alter', 'begin', 'declare', 'create']
@@ -854,41 +854,32 @@ class sqlpyPlus(sqlpython.sqlpython):
         """Displays source code."""
         self._pull(arg, opts)
 
-    def _pull(self, arg, opts, vc=None):        
-        opts.exact = True
+    def _pull(self, arg, opts, vc=None):  
         statekeeper = Statekeeper(opts.dump and self, ('stdout',))
-        (username, gerald_schema) = self.metadata()
         try:
-            for description in self._matching_database_objects(arg, opts):                
-                self.poutput(description.path)
-                txt = description.dbobj.get_ddl()
+            for (owner, object_type, name) in self.current_instance.objects(arg, opts):
+                obj = self.current_instance.object_metadata(owner, object_type, name)
+                txt = obj.get_ddl()
                 if opts.get('lines'):
                     txt = self._with_line_numbers(txt)    
                 if opts.dump:
-                    owner = description.owner or self.current_instance.username
-                    path = os.path.join(owner.lower(), description.type.lower()) \
-                           .replace(' ', '_')
+                    path = os.path.join(owner.lower(), object_type.lower()).replace(' ', '_')
                     try:
                         os.makedirs(path)
                     except OSError:
                         pass
-                    filename = os.path.join(path, '%s.sql' % description.unqualified_name.lower())
+                    filename = os.path.join(path, '%s.sql' % name.lower())
                     self.stdout = open(filename, 'w')
                 if opts.get('num') is not None:
                     txt = txt.splitlines()
                     txt = centeredSlice(txt, center=opts.num+1, width=opts.width)
                     txt = '\n'.join(txt)
                 else:
-                    txt = 'REMARK BEGIN %s\n%s\nREMARK END\n' % (description.path, txt)
-
+                    txt = 'REMARK BEGIN %s/%s/%s\n%s\nREMARK END\n' % (owner, object_type, name, txt)
                 self.poutput(txt)
-                if opts.full:
-                    for dependent_type in ('constraints', 'triggers', 'indexes'):
-                        if hasattr(description.dbobj, dependent_type):
-                            for (depname, depobj) in getattr(description.dbobj, dependent_type).items():
-                                self.poutput('REMARK BEGIN\n%s\nREMARK END\n\n' % depobj.get_ddl())
                 if opts.dump:
                     self.stdout.close()
+                    statekeeper.restore()
                     if vc:
                         subprocess.call(vc + [filename])                    
         except:
@@ -1018,26 +1009,13 @@ class sqlpyPlus(sqlpython.sqlpython):
              ])                    
     def do_find(self, arg, opts):
         """Finds argument in source code or (with -c) in column definitions."""
-              
-        seek = re.compile(self._regex_form_of_search_pattern(arg, exact=opts.col),
-                          re.IGNORECASE)
-        qualified = opts.get('all')
-        for descrip in self._matching_database_objects('*', opts):
-            if opts.col:
-                if hasattr(descrip.dbobj, 'columns'):
-                    for col in descrip.dbobj.columns:
-                        if seek.search(col):
-                            self.poutput('%s.%s' % (m.fullname, col))
-            else:
-                if hasattr(descrip.dbobj, 'source'):
-                    name_printed = False
-                    for (line_num, line) in descrip.dbobj.source:
-                        if seek.search(line):
-                            if not name_printed:
-                                self.poutput(descrip.fullname)
-                                name_printed = True
-                            self.poutput('%d: %s' % (line_num, line))
-            
+        if opts.col:
+            for (owner, object_type, table_name, column_name) in self.current_instance.columns(arg, '%', opts):
+                self.poutput('%s %s.%s.%s' % (object_type, owner, table_name, column_name))
+        else:
+            for (owner, object_type, name, line_number, txt) in self.current_instance.source(arg, opts):
+                self.poutput('%s %s.%s %d: %s' % (object_type, owner, name, line_number, txt))
+           
     def _col_type_descriptor(self, col):
         #if col['type'] in ('integer',):
         #    return col['type']
@@ -1056,28 +1034,31 @@ class sqlpyPlus(sqlpython.sqlpython):
                     #TODO: in postgres, _key_columns returns 'fishies_pkey' instead of 'n'        
         else:
             return []
+
+    standard_options = [
+              all_users_option,
+              make_option('-l', '--long', action='store_true', help='long descriptions'),
+              make_option('-r', '--reverse', action='store_true', help="Reverse order while sorting")]
         
-    @options([all_users_option,
-              make_option('-l', '--long', action='store_true', help='include column #, comments'),
+    @options(standard_options + [
               make_option('-A', '--alpha', action='store_true', help='List columns alphabetically')])
     def do_describe(self, arg, opts):
-        opts.exact = True
         rowlimit = self.rowlimit(arg)
         if opts.alpha:
             sortkey = operator.itemgetter('name')
         else:
             sortkey = operator.itemgetter('sequence')
-        for descrip in self._matching_database_objects(arg, opts):
-            self.tblname = descrip.fullname
+        for (owner, object_type, name) in self.current_instance.objects(arg, opts):
+            obj = self.current_instance.object_metadata(owner, object_type, name)
+            self.tblname = '%s %s.%s' % (object_type, owner, name)
             self.pfeedback(self.tblname)
-            if opts.long and hasattr(descrip.dbobj, 'comments'):
-                if descrip.dbobj.comments:
-                    self.poutput(descrip.dbobj.comments) 
-            if hasattr(descrip.dbobj, 'columns'):
-                cols = sorted(descrip.dbobj.columns.values(), key=sortkey)[:rowlimit]
-                if opts.long and hasattr(descrip.dbobj, 'constraints'):
-                    primary_key_columns = self._key_columns(descrip.dbobj, 'Primary')
-                    unique_key_columns = self._key_columns(descrip.dbobj, 'Unique')
+            if opts.long and hasattr(obj, 'comments') and obj.comments:
+                self.poutput(obj.comments) 
+            if hasattr(obj, 'columns'):
+                cols = sorted(obj.columns.values(), key=sortkey, reverse=bool(opts.reverse))[:rowlimit]
+                if opts.long and hasattr(obj, 'constraints'):
+                    primary_key_columns = self._key_columns(obj, 'Primary')
+                    unique_key_columns = self._key_columns(obj, 'Unique')
                     self.colnames = 'N Name Nullable Type Key Default Comments'.split()
                     self.rows = [(col['sequence'], col['name'], (col['nullable'] and 'NULL') or 'NOT NULL',
                                   self._col_type_descriptor(col), 
@@ -1091,38 +1072,38 @@ class sqlpyPlus(sqlpython.sqlpython):
                                  for col in cols]
                 self.coltypes = [str] * len(self.colnames)
                 self.poutput('%s\n\n' % self.tabular_output(arg.parsed.terminator, self.tblname))
-            elif hasattr(descrip.dbobj, 'increment_by'):
+            elif hasattr(obj, 'increment_by'):
                 self.colnames = 'name min_value max_value increment_by'.split()
                 self.coltypes = [str, int, int, int]
-                self.rows = [(getattr(descrip.dbobj, p) for p in self.colnames)]
+                self.rows = [(getattr(obj, p) for p in self.colnames)]
                 self.poutput('%s\n\n' % self.tabular_output(arg.parsed.terminator, self.tblname))
-            elif hasattr(descrip.dbobj, 'source'):
+            elif hasattr(obj, 'source'):
                 end_heading = re.compile(r'\bDECLARE|BEGIN\b', re.IGNORECASE)
-                for (index, (ln, line)) in enumerate(descrip.dbobj.source):
+                for (index, (ln, line)) in enumerate(obj.source):
                     if end_heading.search(line):
                         break
-                self.poutput(''.join(l for (ln, l) in descrip.dbobj.source[:index]))
-                        
+                self.poutput(''.join(l for (ln, l) in obj.source[:index]))
     @options([all_users_option])            
     def do_deps(self, arg, opts):
-        '''Lists all objects that are dependent upon the object.'''
+        '''Lists indexes, constraints, and triggers depending on an object'''
         #TODO: doesn't account for views; don't know about primary keys
-        for descrip in self._matching_database_objects(arg, opts):
+        for (owner, object_type, name) in self.current_instance.objects(arg, opts):
+            obj = self.current_instance.object_metadata(owner, object_type, name)
             for deptype in ('indexes', 'constraints', 'triggers'):
-                if hasattr(descrip.dbobj, deptype):
-                    for (depname, depobj) in getattr(descrip.dbobj, deptype).items():
+                if hasattr(obj, deptype):
+                    for (depname, depobj) in getattr(obj, deptype).items():
                         self.poutput('%s %s' % (deptype, depname))
                 
     @options([all_users_option])        
     def do_comments(self, arg, opts):
         'Prints comments on a table and its columns.'
-        qualified = opts.get('all')
-        for descrip in self._matching_database_objects(arg, opts):
-            if hasattr(descrip.dbobj, 'comments'):
-                self.poutput(descrip.fullname)
-                self.poutput(descrip.dbobj.comments)
-                if hasattr(descrip.dbobj, 'columns'):
-                    columns = descrip.dbobj.columns.values()
+        for (owner, object_type, name) in self.current_instance.objects(arg, opts):
+            obj = self.current_instance.object_metadata(owner, object_type, name)
+            if hasattr(obj, 'comments'):
+                self.poutput('%s %s.%s' % object_type, owner, name)
+                self.poutput(obj.comments)
+                if hasattr(obj, 'columns'):
+                    columns = obj.columns.values()
                     columns.sort(key=operator.itemgetter('sequence'))
                     for col in columns:
                         comment = col.get('comment')
@@ -1249,11 +1230,7 @@ class sqlpyPlus(sqlpython.sqlpython):
     def _do_dir(self, type, arg, opts):
         self._do_ls("%s/%s%s%s" % (type, str(arg), arg.parsed.terminator, arg.parsed.suffix), opts)
 
-    standard_options = [
-              all_users_option,
-              make_option('-l', '--long', action='store_true', help='long descriptions'),
-              make_option('-r', '--reverse', action='store_true', help="Reverse order while sorting")]
-    
+   
     @options(standard_options)
     def do__dir_tables(self, arg, opts):
         'Shortcut for ``ls table/``'
@@ -1300,12 +1277,12 @@ class sqlpyPlus(sqlpython.sqlpython):
         
     def do__dir_(self, arg, opts, plural_name, str_function):
         long = opts.get('long')
-        reverse = opts.get('reverse') or False
-        for descrip in self._matching_database_objects(arg, opts):
-            if hasattr(descrip.dbobj, plural_name):
-                self.pfeedback('%s on %s' % (plural_name.title(), descrip.fullname))
-                result = [str_function(depobj, long) for depobj in getattr(descrip.dbobj, plural_name).values()]
-                result.sort(reverse=opts.get('reverse') or False)
+        for (owner, object_type, name) in self.current_instance.objects(arg, opts):
+            obj = self.current_instance.object_metadata(owner, object_type, name)
+            if hasattr(obj, plural_name):
+                self.pfeedback('%s on %s' % (plural_name.title(), '%s %s.%s' % (object_type, owner, name)))
+                result = [str_function(depobj, long) for depobj in getattr(obj, plural_name).values()]
+                result.sort(reverse=bool(opts.reverse))
                 self.poutput('\n'.join(result))
         
     @options(standard_options)
@@ -1326,8 +1303,7 @@ class sqlpyPlus(sqlpython.sqlpython):
     @options(standard_options)
     def do__dir_triggers(self, arg, opts):
         '''
-        Called with an exact table name, lists the indexes of that table.
-        Otherwise, acts as shortcut for `ls index/*(arg)*`
+        Called with an exact table name, lists the triggers of that table.
         '''
         self.do__dir_(arg, opts, 'triggers', self._str_trigger)
         
@@ -1531,14 +1507,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         'WINDOW GROUP',         
         'XML SCHEMA')
         
-    def metadata(self):
-        username = self.current_instance.username
-        if self.rdbms == 'oracle':
-            username = username.upper()
-        elif self.rdbms == 'postgres':
-            username = username.lower()
-        return (username, self.current_instance.gerald)
-        
+       
     def _to_sql_wildcards(self, original):
         return original.replace('*','%').replace('?','_')
         #hmm... but these should be escape-able?
@@ -1546,55 +1515,11 @@ class sqlpyPlus(sqlpython.sqlpython):
     def _to_re_wildcards(self, original):
         result = re.escape(original)
         return result.replace('\\*','.*').replace('\\?','.')
-        
-    def _regex_form_of_search_pattern(self, s, exact=False):
-        if not s:
-            return '^[^\.]*$'
-        if '.' in s:
-            seekpatt = r'[/\\]?%s[/\\]?' % (
-                s.replace('*', '.*').replace('?','.').replace('%', '.*'))        
-        else:
-            seekpatt = r'[/\\]?%s[/\\]?' % (
-                s.replace('*', '[^\.]*').replace('?','[^\.]').replace('%', '[^\.]*'))                    
-        if exact:
-            seekpatt = '^%s$' % seekpatt
-        return seekpatt
 
-    def do_refresh(self, arg):
-        '''Refreshes metadata for the specified schema; only required
-           if table structures, etc. have changed.'''
-        if self.current_instance.gerald.complete and self.current_instance.gerald.current:
-            self.current_instance.discover_metadata()
-        else:
-            self.pfeedback('Metadata discovery is already underway.')       
-        
-    def _print_gerald_status_warning(self, gerald_schema):
-        if not gerald_schema.complete:
-            self.pfeedback('Metadata is not available yet - still gathering')
-        elif not gerald_schema.current:
-            self.pfeedback('Metadata is stale - requested refresh still underway')
-                            
-    def _matching_database_objects(self, arg, opts):
-        # doesn't get java$options
-
-        (username, gerald_schema) = self.metadata()                
-        self._print_gerald_status_warning(gerald_schema)
-        if not gerald_schema.complete:
-            raise StopIteration
-        
-        seek = str(arg) and self._regex_form_of_search_pattern(arg, opts.get('exact'))
-        for (name, descrip) in gerald_schema.descriptions.items():
-            if descrip.match_pattern(seek, specific_owner = ((not opts.all) and username)):
-                yield descrip
-   
     def _do_ls(self, arg, opts):
         'Functional core of ``do_ls``, split out into an undecorated version to be callable from other methods'
-        opts.exact = True
-        (username, schemas) = self.metadata()
-        result = [descrip.path for descrip in self._matching_database_objects(arg, opts)]
-        if result:
-            result.sort(reverse=bool(opts.reverse))
-            self.poutput('\n'.join(result))
+        for row in self.current_instance.objects(arg, opts):
+            self.poutput('%s/%s/%s' % row)
                 
     @options(standard_options)
     def do_ls(self, arg, opts):
@@ -1610,8 +1535,8 @@ class sqlpyPlus(sqlpython.sqlpython):
     def do_grep(self, arg, opts):
         """grep {target} {table} [{table2,...}]
         search for {target} in any of {table}'s fields"""    
+        # TODO: permit regex
         arg = self.parsed(arg)
-        opts.exact = True
         args = arg.split()
         if len(args) < 2:
             self.perror(self.do_grep.__doc__)
@@ -1626,25 +1551,26 @@ class sqlpyPlus(sqlpython.sqlpython):
         re_pattern = re.compile(self._to_re_wildcards(pattern), 
                                 (opts.ignorecase and re.IGNORECASE) or 0)
         for target in targets:
-            for descrip in self._matching_database_objects(target, opts):
-                self.pfeedback(descrip.fullname)
-                if hasattr(descrip.dbobj, 'columns'):
+            for (owner, object_type, name) in self.current_instance.objects(target, opts):
+                obj = self.current_instance.object_metadata(owner, object_type, name)
+                self.pfeedback('%s %s.%s' % (object_type, owner, name))
+                if hasattr(obj, 'columns'):
                     clauses = []
-                    for col in descrip.dbobj.columns:
+                    for col in obj.columns:
                         clauses.append(comparitor % (col, sql_pattern))
-                    sql = "SELECT * FROM %s WHERE 1=0\n%s;" % (descrip.fullname, ' '.join(clauses))
+                    sql = "SELECT * FROM %s.%s WHERE 1=0\n%s;" % (owner, name, ' '.join(clauses))
                     sql = self.parsed(sql, terminator=arg.parsed.terminator or ';',
                                       suffix=arg.parsed.suffix)
                     self.do_select(sql)
-                elif hasattr(descrip.dbobj, 'source'):
-                    for (line_num, line) in descrip.dbobj.source:
+                elif hasattr(obj, 'source'):
+                    for (line_num, line) in obj.source:
                         if re_pattern.search(line):
                             self.poutput('%4d: %s' % (line_num, line))
                 
 
     def _cast_as_char(self):
         'self._cast_as_char() => Returns the RDBMS-equivalent "CAST(%s AS VARCHAR) expression.' 
-        converter = {'oracle': 'TO_CHAR(%s)'}.get(self.rdbms, 'CAST(%s AS VARCHAR)')
+        converter = {'oracle': 'TO_CHAR(%s)', 'mysql': 'CAST(%s AS CHAR)'}.get(self.rdbms, 'CAST(%s AS VARCHAR)')
         return converter
     
     def _execute(self, sql, bindvars={}):
