@@ -10,7 +10,7 @@
 
 import cmd2,getpass,binascii,re,os,platform
 import pyparsing, connections
-__version__ = '1.7.0'
+__version__ = '1.7.2'
 try:
     import cx_Oracle
 except ImportError:
@@ -83,7 +83,11 @@ class sqlpython(cmd2.Cmd):
             self.list_instances()
             return
         if self.commit_on_exit:
-            instance.connection.commit()
+            try:
+                instance.connection.commit()
+            except Exception, e:
+                self.perror('Error while committing:')
+                self.perror(str(e))                
         self.instances.pop(instance_number)
         if instance_number == self.instance_number:
             self.no_instance()
@@ -116,6 +120,7 @@ class sqlpython(cmd2.Cmd):
                                     help='close connection {N} (or current)'),
                    cmd2.make_option('-C', '--closeall', action='store_true',
                                     help='close all connections'),
+                   cmd2.make_option('--postgresql', action='store_true', help='Connect to postgreSQL: `connect --postgresql [DBNAME [USERNAME]]`'),
                    cmd2.make_option('--postgres', action='store_true', help='Connect to postgreSQL: `connect --postgres [DBNAME [USERNAME]]`'),
                    cmd2.make_option('--oracle', action='store_true', help='Connect to an Oracle database'),
                    cmd2.make_option('--mysql', action='store_true', help='Connect to a MySQL database'),
@@ -138,6 +143,7 @@ class sqlpython(cmd2.Cmd):
         connect user/password@SID  (Oracle is the default RDBMS target)
         connect --postgres --hostname=hostname dbname username
         connect --mysql dbname username'''
+        opts.postgres = opts.postgres or opts.postgresql
         if opts.closeall:
             self.closeall()
             return 
@@ -154,8 +160,9 @@ class sqlpython(cmd2.Cmd):
 
         try:
             db_instance = connections.DatabaseInstance(arg, opts, default_rdbms = self.default_rdbms)
-        except:
+        except Exception, e:
             self.perror('Connection failure.\n' + self.do_connect.__doc__)
+            self.perror(str(e))
             return
         if opts.add or (self.instance_number is None):
             try:
